//cluster simian volume preprocessor, part 1.
//DeMarle 06/17/2004
//based on bricker.cpp by Mark Hartner and genVGH.cpp by Joe Kniss

//This program reads in a scalar volume and prepares it for use by 
//cluster simian. The preparation involves computing derivatives volumes,
//quantizing the data, blurring the normals, and padding to an even power of 
//two size. This program operates in a slice by slice fashion from the input 
//files, so that it doesn't need to keep the whole volume in memory, which 
//makes it suitable for large files.

//To finish the preparation process, run vol_prep_part2, which merges the 
//computed volumes, and then breaks the merged volume into sub blocks.

#include <stdio.h> //printf, scanf
#include <unistd.h> //read
#include <fcntl.h>  //O_LARGEFILE
#include <math.h> //sqrt
#include <stdlib.h> //exit
#include <string.h> //memset, memcpy

//integral dimensions of volume data in memory space
int izd, iyd, ixd;
int iydXixd = 0;
//padded onto power of two
int pzd, pyd, pxd; 
int pydXpxd = 0;
//dimensions of a subblock within the padded volume
int bzd, byd, bxd;

int readslice;
float *gradV3[3];
float *gstencil[3];
unsigned char *c_v_in_slice[3];
unsigned char *dstencil[3];

//raw input data extremes
unsigned char dmax = 0;
unsigned char dmin = 255;
//gradient magnitude extremes
float gmmax = -100000000;
float gmmin = 100000000;
//hessian extremes
float hmax = -100000000;
float hmin = 100000000;
#ifndef MAX
#define MAX(x,y) (((x)>(y)) ? (x) : (y))
#endif
#ifndef MIN
#define MIN(x,y) (((x)<(y)) ? (x) : (y))
#endif
inline double affine(const double i, const double x, const double I, 
		     const double o, const double O)
{
   return  ( ((O)-(o))*((x)-(i)) / ((I)-(i)) + (o) );
}

inline void normalizeV3(float v[3])
{
  float len = (float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]); 
  if(len > 0){
    v[0] /= len;
    v[1] /= len;
    v[2] /= len;
  }
}

inline void csaddV3(float out[3], float s, float in[3])
{
  out[0] += s*in[0];
  out[1] += s*in[1];
  out[2] += s*in[2];
}

#define MAXNAMELEN 256
char filename[MAXNAMELEN]; //name of raw data file
char ifilename[MAXNAMELEN]; //name for transient an input file
char ofilename[MAXNAMELEN]; //name for transient an output file

int c_v_in_fd; //unsigned character value (aka data aka scalar) input file descriptor

int f_gm_fd;  //floating point gradient magnitude file descriptor
int f_h_fd;  //floating point hessian file descriptor
int f_g_fd; //floating point gradient file descriptor

int bf_g_fd; //blurred floating point gradient file descriptor

int q_v_fd; //quantized (aka character) value file descriptor
int q_gm_fd;  //quantized gradient magnitude file descriptor
int q_h_fd;  //quantized hessian file descriptor
int q_g_fd;  //quantized gradient file descriptor

int pq_v_fd; //padded quantized value file descriptor
int pq_gm_fd;  //padded quantized gradient magnitude file descriptor
int pq_h_fd;  //padded quantized hessian file descriptor
int pq_g_fd;  //padded quantized gradient file descriptor

int rc; //system call return code

int compute_derivatives(int slice, 
			unsigned char *d[3],
			float *gstencil[3],
			float *gmag,
			float *hess
			) {

  //compute gradient and gradient magnitude
  
  int &i = slice;
  float *gradV3 = gstencil[1];

  //printf("Computing gradient and gradient magnitude fields\n");
  for(int j = 0; j < (iyd); ++j) {
    for(int k = 0; k < (ixd); ++k) {
      if((k<1)||(k>ixd-2)||(j<1)||(j>iyd-2)||(i<1)||(i>izd-2)) {
	//on border cells, gradient vector length and gradient magnitude are set to zero
	// use * 3 because each grad vector takes up three floats
	gradV3[j*ixd*3 + k*3 + 0] = 0;
	gradV3[j*ixd*3 + k*3 + 1] = 0;
	gradV3[j*ixd*3 + k*3 + 2] = 0;
	gmag[j*ixd + k] = 0;
      }
      else {
	//inside gradient vector components are computed as difference of two neighbor cell values
	float dx = (float)(d[1][j*ixd     +(k+1)] - d[1][j*ixd     + (k-1)]);
	float dy = (float)(d[1][(j+1)*ixd + k   ] - d[1][(j-1)*ixd + k    ]);
	float dz = (float)(d[2][j*ixd     + k   ] - d[0][j*ixd     + k    ]);
	
	//use * 3 because each grad vector takes up three floats
	gradV3[j*ixd*3 + k*3 + 0] = dx;
	gradV3[j*ixd*3 + k*3 + 1] = dy;
	gradV3[j*ixd*3 + k*3 + 2] = dz;
	
	//gradient magnitude is just the sqrt of the sum of squares
	gmag[j*ixd + k] = (float)sqrt(dx*dx+dy*dy+dz*dz);
	
	//remember the min and max grad magnitude
	gmmax = MAX(gmag[j*ixd + k], gmmax);
	gmmin = MIN(gmag[j*ixd + k], gmmin);
	
	//remember the min and max data magnitude
	dmax = MAX(d[1][j*ixd +k], dmax);
	dmin = MIN(d[1][j*ixd +k], dmin);
      }
    }
  }

  rc = write(f_g_fd, (void*)gradV3, iydXixd*3*sizeof(float));
  if (rc != iydXixd*3*sizeof(float)) {
    return -1;
  }
  rc = write(f_gm_fd, (void*)gmag, iydXixd*sizeof(float));
  if (rc != iydXixd*sizeof(float)) {
    return -1;
  }


  //compute hessian
  //compute the 2nd derivative in the gradient direction
  //printf("Computing 2nd derivative in the gradient direction\n");
  for(int j = 0; j < (iyd); ++j){
    for(int k = 0; k < (ixd); ++k){
      
      
      if((k<1)||(k>ixd-2)||(j<1)||(j>iyd-2)||(i<1)||(i>izd-2)){
	//boundary cell values are again set to zero
	hess[j*ixd + k] = 0;
      }
      else {
	// each set of three is like the above dx,dy,dz calc
	// but each set adds 1 which accounts for the x, y and z coefficients in the vector storage
	
	float h[9];
	//gradient of the +0=x gradient cooeficient
	h[0] = gstencil[1][j*ixd*3 + (k+1)*3 + 0] - gstencil[1][j*ixd*3 + (k-1)*3 + 0];
	h[1] = gstencil[1][(j+1)*ixd*3 + k*3 + 0] - gstencil[1][(j-1)*ixd*3 + k*3 + 0];
	h[2] = gstencil[2][j*ixd*3 + k*3 + 0] - gstencil[0][j*ixd*3 + k*3 + 0];
	
	//gradient of the +1=y gradient cooeficient
	h[3] = gstencil[1][j*ixd*3 + (k+1)*3 + 1] - gstencil[1][j*ixd*3 + (k-1)*3 + 1];
	h[4] = gstencil[1][(j+1)*ixd*3 + k*3 + 1] - gstencil[1][(j-1)*ixd*3 + k*3 + 1];
	h[5] = gstencil[2][j*ixd*3 + k*3 + 1] - gstencil[0][j*ixd*3 + k*3 + 1];

	//gradient of the +2=z gradient cooeficient
	h[6] = gstencil[1][j*ixd*3 + (k+1)*3 + 2] - gstencil[1][j*ixd*3 + (k-1)*3 + 2];
	h[7] = gstencil[1][(j+1)*ixd*3 + k*3 + 2] - gstencil[1][(j-1)*ixd*3 + k*3 + 2];
	h[8] = gstencil[2][j*ixd*3 + k*3 + 2] - gstencil[0][j*ixd*3 + k*3 + 2];


	float tv[3];
	//x value of tg, is x coeeficient(+0) of gradient at this cell divided by grad magnitude at this cell
	//y value of tg, is y coeeficient(+1) of gradient at this cell divided by grad magnitude at this cell
	//z value of tg, is z coeeficient(+2) of gradient at this cell divided by grad magnitude at this cell					
	float tg[3] = {gradV3[j*ixd*3 + k*3 + 0]/gmag[j*ixd + k], 
		       gradV3[j*ixd*3 + k*3 + 1]/gmag[j*ixd + k],
		       gradV3[j*ixd*3 + k*3 + 2]/gmag[j*ixd + k]};

	tv[0] = tg[0] * h[0] + 
	  tg[1] * h[1] +
	  tg[2] * h[2];
	tv[1] = tg[0] * h[3] + 
	  tg[1] + /*h[4] + ?why is hyy not included?*/
	  tg[2] * h[5];
	tv[2] = tg[0] * h[6] + 
	  tg[1] * h[7] +
	  tg[2] * h[8];
	
	//finally compute the hessian magnitude
	hess[j*ixd + k] = tg[0] * tv[0] + 
	  tg[1] * tv[1] +
	  tg[2] * tv[2];
	
	//remember min and max hessian values
	hmax = MAX(hess[j*ixd + k], hmax);
	hmin = MIN(hess[j*ixd + k], hmin);	
      }
    }
  }


  rc = write(f_h_fd, (void*)hess, iydXixd*sizeof(float));
  if (rc != iydXixd*sizeof(float)) {
    return -1;
  }

  return 0;
}


int blurV3D(int slice, 
	    float *gin,
	    float *gout[3],
	    float w0, float w1, float w2, float w3
	    ) {
    
  //zero newest slice//////////////////////////
  memset(gout[2], 0, iydXixd*3*sizeof(float));

  //merge in just read slice into all three slices//////////////////
  //memcpy(gout[1], gin[1], iydXixd*3*sizeof(float)); //DEBUG

  if (slice == 0) {
    memset(gout[1], 0, iydXixd*3*sizeof(float));
  } else {

    for(int j = 1; j<iyd-1; ++j){
      for(int k = 1; k<ixd-1; ++k){
	int inloc = (j*ixd + k)*3;      
	
	csaddV3(&gout[0][((j-1)*ixd + (k-1))*3], w3, &gin[inloc]);
	csaddV3(&gout[0][((j-1)*ixd + (k+0))*3], w2, &gin[inloc]);
	csaddV3(&gout[0][((j-1)*ixd + (k+1))*3], w3, &gin[inloc]);
	csaddV3(&gout[0][((j+0)*ixd + (k-1))*3], w2, &gin[inloc]);
	csaddV3(&gout[0][((j+0)*ixd + (k+0))*3], w1, &gin[inloc]);
	csaddV3(&gout[0][((j+0)*ixd + (k+1))*3], w2, &gin[inloc]);
	csaddV3(&gout[0][((j+1)*ixd + (k-1))*3], w3, &gin[inloc]);
	csaddV3(&gout[0][((j+1)*ixd + (k+0))*3], w2, &gin[inloc]);
	csaddV3(&gout[0][((j+1)*ixd + (k+1))*3], w3, &gin[inloc]);
	
	csaddV3(&gout[1][((j-1)*ixd + (k-1))*3], w2, &gin[inloc]);
	csaddV3(&gout[1][((j-1)*ixd + (k+0))*3], w1, &gin[inloc]);
	csaddV3(&gout[1][((j-1)*ixd + (k+1))*3], w2, &gin[inloc]);
	csaddV3(&gout[1][((j+0)*ixd + (k-1))*3], w1, &gin[inloc]);
	csaddV3(&gout[1][((j+0)*ixd + (k+0))*3], w0, &gin[inloc]);
	csaddV3(&gout[1][((j+0)*ixd + (k+1))*3], w1, &gin[inloc]);
	csaddV3(&gout[1][((j+1)*ixd + (k-1))*3], w2, &gin[inloc]);
	csaddV3(&gout[1][((j+1)*ixd + (k+0))*3], w1, &gin[inloc]);
	csaddV3(&gout[1][((j+1)*ixd + (k+1))*3], w2, &gin[inloc]);
	
	csaddV3(&gout[2][((j-1)*ixd + (k-1))*3], w3, &gin[inloc]);
	csaddV3(&gout[2][((j-1)*ixd + (k+0))*3], w2, &gin[inloc]);
	csaddV3(&gout[2][((j-1)*ixd + (k+1))*3], w3, &gin[inloc]);
	csaddV3(&gout[2][((j+0)*ixd + (k-1))*3], w2, &gin[inloc]);
	csaddV3(&gout[2][((j+0)*ixd + (k+0))*3], w1, &gin[inloc]);
	csaddV3(&gout[2][((j+0)*ixd + (k+1))*3], w2, &gin[inloc]);
	csaddV3(&gout[2][((j+1)*ixd + (k-1))*3], w3, &gin[inloc]);
	csaddV3(&gout[2][((j+1)*ixd + (k+0))*3], w2, &gin[inloc]);
	csaddV3(&gout[2][((j+1)*ixd + (k+1))*3], w3, &gin[inloc]);
      }
    }
  }

  //write out oldest slice
  //printf("writing slice %d\n", slice);

  if (slice > 0) {
    rc = write(bf_g_fd, (void*)gout[0], iydXixd*3*sizeof(float));
    if (rc != iydXixd*3*sizeof(float)) {
      return -1;
    }
  }

  if (slice == izd-1) {
    memset(gout[1], 0, iydXixd*3*sizeof(float));
    rc = write(bf_g_fd, (void*)gout[1], iydXixd*3*sizeof(float));
    if (rc != iydXixd*3*sizeof(float)) {
      return -1;
    }
  }

  return 0;
}

int isPow2(int size)
{
  double tmp = size/2.0;
  while(tmp > 1.0){
    tmp = tmp/2.0;
  }
  if(tmp != 1.0) return 0;
  else return 1;
}

int getPow2(int sz)
{
  double tmp = sz/2.0;
  int i = 2;
  while(tmp > 1.0){
    i *= 2;
    tmp = tmp/2.0;
  }
  return i;
}


int main(int argc, char **argv) {

  //STEP 1////////////////////////////////////////////////////////////////
  ///ASKING FOR INPUT VOLUME STATS

  printf("What is the input file name? ");
  rc = scanf("%s", filename);
  printf("filename : %s\n", filename);

  printf("What are the integral dimensions (in x y z format where x changes fastest)? ");
  scanf("%d %d %d", &ixd, &iyd, &izd);
  printf("dimensions : x=%d, y=%d, z=%d\n", ixd, iyd, izd);

  iydXixd = iyd*ixd;
  c_v_in_slice[0] = new unsigned char[iydXixd];
  c_v_in_slice[1] = new unsigned char[iydXixd];
  c_v_in_slice[2] = new unsigned char[iydXixd];

  float *gmag = new float[iydXixd];
  float *hess = new float[iydXixd];

  gradV3[0] = new float[iydXixd*3];
  gradV3[1] = new float[iydXixd*3];
  gradV3[2] = new float[iydXixd*3];
  
  //STEP 2//////////////////////////////////////////////////////////////////
  ///COMPUTING DERIVATIVE VOLUMES
  printf("Computing gradient, gradient magnitude and hessian volumes.\n");

  //INPUT FILES
  c_v_in_fd = open(filename, O_RDONLY|O_LARGEFILE);
  if (c_v_in_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }

  //OUTPUT FILES
  sprintf(ofilename, "%s.fgm_tmp", filename);
  f_gm_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (f_gm_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }
  sprintf(ofilename, "%s.fh_tmp", filename);
  f_h_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (f_h_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }
  sprintf(ofilename, "%s.fg_tmp", filename);
  f_g_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (f_g_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }

  readslice = 0;
  //read first slice
  rc = read(c_v_in_fd, (unsigned char*)c_v_in_slice[readslice%3], iydXixd*sizeof(unsigned char));
  if (rc != iydXixd*sizeof(unsigned char)) {
    printf("Problem reading input file.\n");
    return -1;
  }
  printf("Just read %d\n", readslice);

  readslice ++;
  while (readslice <= izd) {
    if (readslice != izd) {
      rc = read(c_v_in_fd, (unsigned char*)c_v_in_slice[readslice%3], iydXixd*sizeof(unsigned char));
      if (rc != iydXixd*sizeof(unsigned char)) {
	printf("Problem reading input file.\n");
	return -1;
      }
    }

    int slice2xago = (readslice-2)%3;
    int slice1xago = (readslice-1)%3;
    int slicejustr = (readslice-0)%3;

    if (!(readslice % 10)) {
      if (readslice != izd)
	printf("Just read %d\t", readslice);
      else 
	printf("\t\t");
      printf("Computing %d\t", readslice-1);
      printf("2xago = %d\t", slice2xago);
      printf("1xago = %d\t", slice1xago);
      printf("justr = %d\n", slicejustr);
    }

    dstencil[0] = c_v_in_slice[slice2xago];
    dstencil[1] = c_v_in_slice[slice1xago];
    dstencil[2] = c_v_in_slice[slicejustr];
    
    gstencil[0] = gradV3[slice2xago];
    gstencil[1] = gradV3[slice1xago];
    gstencil[2] = gradV3[slicejustr];

    rc = compute_derivatives(readslice-1, dstencil, gstencil, gmag, hess);
    if (rc) {
      printf("Problem computing derivatives.\n");
    }
    readslice++;
  }

  close(c_v_in_fd);
  close(f_g_fd);
  close(f_gm_fd);
  close(f_h_fd);

  //STEP 3//////////////////////////////////////////////////////////////////
  ///BLURRING NORMALS
  printf("Blurring normals.\n");

  //INPUT FILES
  sprintf(ofilename, "%s.fg_tmp", filename);
  f_g_fd = open(ofilename, O_RDONLY|O_LARGEFILE);
  if (f_g_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }

  //OUTPUT FILES
  sprintf(ofilename, "%s.bfg_tmp", filename);
  bf_g_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (bf_g_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }

  //reuse these from above
  //float *gradV3[3];
  //gradV3[0] = new float[iydXixd*3];
  //gradV3[1] = new float[iydXixd*3];
  //gradV3[2] = new float[iydXixd*3];
  //float *gstencil[3];

  float *bgradV3[3];
  bgradV3[0] = new float[iydXixd*3];
  bgradV3[1] = new float[iydXixd*3];
  bgradV3[2] = new float[iydXixd*3];
  float *bgstencil[3];

  readslice = 0;
  //read first slice
  rc = read(f_g_fd, (void*)gradV3[readslice%3], iydXixd*3*sizeof(float));
  if (rc != iydXixd*3*sizeof(float)) {
    printf("Problem reading input file.\n");
    return -1;
  }
  printf("Just read %d\n", readslice);

  float w0 = 1.0;
  float w1 = 0.3;
  float w2 = 0.2;
  float w3 = 0.1;
  float div = w0 + 6*w1 + 12*w2 + 8*w3;
  w0 = w0 / div;
  w1 = w1 / div;
  w2 = w2 / div;
  w3 = w3 / div;

  readslice ++;
  while (readslice <= izd) {
    if (readslice != izd) {
      rc = read(f_g_fd, (void*)gradV3[readslice%3], iydXixd*3*sizeof(float));
      if (rc != iydXixd*3*sizeof(float)) {
	printf("Problem reading input file.\n");
	break;
      }
    }

    int slice2xago = (readslice-2)%3;
    int slice1xago = (readslice-1)%3;
    int slicejustr = (readslice-0)%3;

    if (!(readslice % 10)) {
      if (readslice != izd)
	printf("Just read %d\t", readslice);
      else 
	printf("\t\t");
      printf("Computing %d\t", readslice-1);
      printf("2xago = %d\t", slice2xago);
      printf("1xago = %d\t", slice1xago);
      printf("justr = %d\n", slicejustr);
    }
    
    gstencil[0] = gradV3[slice2xago];
    gstencil[1] = gradV3[slice1xago];
    gstencil[2] = gradV3[slicejustr];

    bgstencil[0] = bgradV3[slice2xago];
    bgstencil[1] = bgradV3[slice1xago];
    bgstencil[2] = bgradV3[slicejustr];

    rc = blurV3D(readslice-1, gstencil[1], bgstencil, w0, w1, w2, w3);
    if (rc) {
      printf("Problem blurring normals.\n");
      return(-1);
    }
    readslice++;
  }

  close(f_g_fd);
  close(bf_g_fd);

  //STEP 4//////////////////////////////////////////////////////////////////
  ///QUANTIZING FLOATING POINT VOLUMES TO UNSIGNED CHAR
  printf("Quantizing into 8 bits.\n");

  printf("dmin %d dmax %d\n", dmin, dmax);
  printf("gmmin %f gmmax %f\n", gmmin, gmmax);
  printf("hmin %f hmax %f\n", hmin, hmax);

  //INPUT FILES
  c_v_in_fd = open(filename, O_RDONLY|O_LARGEFILE);
  if (c_v_in_fd <= 0) {
    printf("Problem opening input file.\n");
   return -1;
  } 
  sprintf(ifilename, "%s.fgm_tmp", filename);
  f_gm_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
  if (f_gm_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }
  sprintf(ifilename, "%s.fh_tmp", filename);
  f_h_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
  if (f_h_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }
  sprintf(ifilename, "%s.bfg_tmp", filename);
  bf_g_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
  if (bf_g_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }

  //OUTPUT FILES
  sprintf(ofilename, "%s.qv_tmp", filename);
  q_v_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (q_v_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }
  sprintf(ofilename, "%s.qgm_tmp", filename);
  q_gm_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (q_gm_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }
  sprintf(ofilename, "%s.qh_tmp", filename);
  q_h_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (q_h_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }
  sprintf(ofilename, "%s.qg_tmp", filename);
  q_g_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (q_g_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }

  unsigned char *d = c_v_in_slice[0];
  float *g = gstencil[0];

  unsigned char od[iydXixd];
  unsigned char ogm[iydXixd];
  unsigned char oh[iydXixd];
  unsigned char og[iydXixd*3];

  readslice = 0;
  while (readslice < izd) {
    if (!(readslice % 10))
      printf("quantizing slice %d\n", readslice);

    rc = read(c_v_in_fd, (unsigned char*)d, iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem reading input file.\n");
      return -1;
    }
    rc = read(f_gm_fd, (float*)gmag, iydXixd*sizeof(float));
    if (rc != iydXixd*sizeof(float)) {
      printf("Problem reading input file.\n");
      return -1;
    }
    rc = read(f_h_fd, (float*)hess, iydXixd*sizeof(float));
    if (rc != iydXixd*sizeof(float)) {
      printf("Problem reading input file.\n");
      return -1;
    }
    rc = read(bf_g_fd, (float*)g, iydXixd*3*sizeof(float));
    if (rc != iydXixd*3*sizeof(float)) {
      printf("Problem reading input file.\n");
      return -1;
    }

    int i = readslice;
    for(int j = 0; j < (iyd); ++j){
      for(int k = 0; k < (ixd); ++k) {
	if((k<1)||(k>ixd-2)||(j<1)||(j>iyd-2)||(i<1)||(i>izd-2)) {
	  //boundary cells are set to zero
	  od[j*ixd + k] = 0;
	  ogm[j*ixd + k] = 0;
	  oh[j*ixd + k] = 0;	
	  og[j*ixd*3 + k*3 + 0] = 0;
	  og[j*ixd*3 + k*3 + 1] = 0;
	  og[j*ixd*3 + k*3 + 2] = 0;
	} else {

	  od[j*ixd + k] = (unsigned char)affine(dmin, d[j*ixd + k], dmax, 0, 255);

	  ogm[j*ixd + k] = (unsigned char)affine(gmmin, gmag[j*ixd + k], gmmax, 0, 255);
	  
	  if(hess[j*ixd + k] < 0){
	    float th = (float)affine(hmin, hess[j*ixd + k], 0, 0, 1);
	    oh[j*ixd + k] = (unsigned char)affine(0,th,1, 0, 255/3); 
	  } else {
	    float th = (float)affine(0, hess[j*ixd + k], hmax, 0, 1);
	    oh[j*ixd + k] = (unsigned char)affine(0,th,1,255/3,255/3*2);
	  }

	  normalizeV3(&g[j*ixd*3 + k*3 + 0]);
	  og[j*ixd*3 + k*3 + 0] = (unsigned char)(g[j*ixd*3 + k*3 + 0]*128 + 128);
	  og[j*ixd*3 + k*3 + 1] = (unsigned char)(g[j*ixd*3 + k*3 + 1]*128 + 128);
	  og[j*ixd*3 + k*3 + 2] = (unsigned char)(g[j*ixd*3 + k*3 + 2]*128 + 128);
	}
      }
    }

    rc = write(q_v_fd, (void*)od, iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem writing output file.\n");
      return -1;
    }
    rc = write(q_gm_fd, (void*)ogm, iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem writing output file.\n");
      return -1;
    }
    rc = write(q_h_fd, (void*)oh, iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem writing output file.\n");
      return -1;
    }
    rc = write(q_g_fd, (void*)og, iydXixd*3*sizeof(unsigned char));
    if (rc != iydXixd*3*sizeof(unsigned char)) {
      printf("Problem writing output file.\n");
      return -1;
    }

    readslice++;
  }

  close(c_v_in_fd);
  close(f_gm_fd);
  close(f_h_fd);
  close(bf_g_fd);
  close(q_v_fd);
  close(q_g_fd);
  close(q_gm_fd);
  close(q_h_fd);

  //STEP 5//////////////////////////////////////////////////////////////////
  ///PADDING TO POWER OF 2 SIZE
  printf("Padding to a power of two size.\n");
  
  pzd = getPow2(izd);
  pyd = getPow2(iyd);
  pxd = getPow2(ixd);

  pydXpxd = pyd*pxd;

  if(!isPow2(izd) || !isPow2(iyd) || !isPow2(ixd)) {


    //INPUT FILES
    sprintf(ifilename, "%s.qv_tmp", filename);
    q_v_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (q_v_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }
    sprintf(ifilename, "%s.qgm_tmp", filename);
    q_gm_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (q_gm_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }
    sprintf(ifilename, "%s.qh_tmp", filename);
    q_h_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (q_h_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }
    sprintf(ifilename, "%s.qg_tmp", filename);
    q_g_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (q_g_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }

    //OUTPUT FILES
    sprintf(ofilename, "%s.pqv_tmp", filename);
    pq_v_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
    if (pq_v_fd <= 0) {
      printf("Problem opening output file.\n");
      return -1;
    }
    sprintf(ofilename, "%s.pqgm_tmp", filename);
    pq_gm_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
    if (pq_gm_fd <= 0) {
      printf("Problem opening output file.\n");
      return -1;
    }
    sprintf(ofilename, "%s.pqh_tmp", filename);
    pq_h_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
    if (pq_h_fd <= 0) {
      printf("Problem opening output file.\n");
      return -1;
    }
    sprintf(ofilename, "%s.pqg_tmp", filename);
    pq_g_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
    if (pq_g_fd <= 0) {
      printf("Problem opening output file.\n");
      return -1;
    }

        
    printf("old sizes = %d %d %d\n", izd, iyd, ixd);
    printf("new sizes = %d %d %d\n", pzd, pyd, pxd);

    unsigned char dataI[iydXixd];
    unsigned char gmagI[iydXixd];
    unsigned char hessI[iydXixd];
    unsigned char gradI[iydXixd*3];

    unsigned char dataO[pydXpxd];
    unsigned char gmagO[pydXpxd];
    unsigned char hessO[pydXpxd];
    unsigned char gradO[pydXpxd*3];
	
    for(int i=0; i<pzd; ++i) {

      if (i < izd) {
	rc = read(q_v_fd, dataI, iydXixd*sizeof(unsigned char));
	if (rc != iydXixd*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
	rc = read(q_gm_fd, gmagI, iydXixd*sizeof(unsigned char));
	if (rc != iydXixd*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
	rc = read(q_h_fd, hessI, iydXixd*sizeof(unsigned char));
	if (rc != iydXixd*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
	rc = read(q_g_fd, gradI, iydXixd*3*sizeof(unsigned char));
	if (rc != iydXixd*3*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
      }

      for(int j=0; j<pyd; ++j) {
	for(int k=0; k<pxd; ++k) {
	  if ((i < izd) && (j < iyd) & (k < ixd)) {
	    dataO[j*pxd + k] = dataI[j*ixd + k];
	    gmagO[j*pxd + k] = gmagI[j*ixd + k];
	    hessO[j*pxd + k] = hessI[j*ixd + k];
	    gradO[j*pxd*3 + k*3 + 0] = gradI[j*ixd*3 + k*3 + 0];	
	    gradO[j*pxd*3 + k*3 + 1] = gradI[j*ixd*3 + k*3 + 1];	
	    gradO[j*pxd*3 + k*3 + 2] = gradI[j*ixd*3 + k*3 + 2];	
	  } else {
	    dataO[j*pxd + k] = 0;
	    gmagO[j*pxd + k] = 0;
	    hessO[j*pxd + k] = 0;
	    gradO[j*pxd*3 + k*3 + 0] = 0;
	    gradO[j*pxd*3 + k*3 + 1] = 0;
	    gradO[j*pxd*3 + k*3 + 2] = 0;
	  }
	}
      }

      rc = write(pq_v_fd, (void*)dataO, pydXpxd*sizeof(unsigned char));
      if (rc != pydXpxd*sizeof(unsigned char)) {
	printf("Problem writing output file.\n");
	return -1;
      }
      rc = write(pq_gm_fd, (void*)gmagO, pydXpxd*sizeof(unsigned char));
      if (rc != pydXpxd*sizeof(unsigned char)) {
	printf("Problem writing output file.\n");
	return -1;
      }
      rc = write(pq_h_fd, (void*)hessO, pydXpxd*sizeof(unsigned char));
      if (rc != pydXpxd*sizeof(unsigned char)) {
	printf("Problem writing output file.\n");
	return -1;
      }
      rc = write(pq_g_fd, (void*)gradO, pydXpxd*3*sizeof(unsigned char));
      if (rc != pydXpxd*3*sizeof(unsigned char)) {
	printf("Problem writing output file.\n");
	return -1;
      }     
    }

    close(pq_v_fd);
    close(pq_g_fd);
    close(pq_gm_fd);
    close(pq_h_fd);

  } else {

    printf("Already a power of 2 size, just copying deriv volumes.\n");
    sprintf(ifilename, "%s.qv_tmp", filename);
    sprintf(ofilename, "%s.pqv_tmp", filename);
    char cmd[2*MAXNAMELEN + 8];
    sprintf(cmd, "ln -s %s %s\n", ifilename, ofilename);
    system(cmd);

    sprintf(ifilename, "%s.qgm_tmp", filename);
    sprintf(ofilename, "%s.pqgm_tmp", filename);
    sprintf(cmd, "ln -s %s %s\n", ifilename, ofilename);
    system(cmd);


    sprintf(ifilename, "%s.qh_tmp", filename);
    sprintf(ofilename, "%s.pqh_tmp", filename);
    sprintf(cmd, "ln -s %s %s\n", ifilename, ofilename);
    system(cmd);

    sprintf(ifilename, "%s.qg_tmp", filename);
    sprintf(ofilename, "%s.pqg_tmp", filename);
    sprintf(cmd, "ln -s %s %s\n", ifilename, ofilename);
    system(cmd);
  }

  printf("Now run vol_prep_part2 to merge the computed volumes and break the result into sub blocks for cluster simian.\n");
  printf("The prepared volumes sizes are now x=%d y=%d z=%d\n", pxd, pyd, pzd);
}
